/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997-2006 Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Sam Lantinga
    slouken at libsdl.or
*/
#include "SDL_config.h"

// Public includes.
#include "SDL_timer.h"

// Audio internal includes.
#include "SDL_audio.h"
#include "../SDL_audiomem.h"
#include "../SDL_sysaudio.h"
#include "../SDL_audio_c.h"

// Wii audio internal includes.
#include <ogcsys.h>
#include "SDL_wiiaudio.h"

#define SAMPLES_PER_DMA_BUFFER (512)
#define DMA_BUFFER_SIZE (SAMPLES_PER_DMA_BUFFER*sizeof(Uint32))

static const char WIIAUD_DRIVER_NAME[] = "wii";
static Uint32 dma_buffers[2][SAMPLES_PER_DMA_BUFFER<<1] __attribute__((aligned(32)));
static Uint32 dma_buffers_size[2] = {DMA_BUFFER_SIZE, DMA_BUFFER_SIZE};
static Uint8 whichab = 0;

#define AUDIOSTACK 8192
static lwp_t athread = LWP_THREAD_NULL;
static Uint8 astack[AUDIOSTACK];
static bool stopaudio = false;

/****************************************************************************
 * Audio Threading
 ***************************************************************************/
static void *
AudioThread (void *arg)
{
	while (1)
	{
		if(stopaudio)
			break;

		memset(dma_buffers[whichab], 0, DMA_BUFFER_SIZE);
		dma_buffers_size[whichab] = DMA_BUFFER_SIZE;

		// Is the device ready?
		if (current_audio && !current_audio->paused)
		{
			SDL_LockMutex(current_audio->mixer_lock);

			if (current_audio->convert.needed)
			{
				// Get the client to produce audio
				current_audio->spec.callback(
						current_audio->spec.userdata,
						current_audio->convert.buf,
						current_audio->convert.len);

				// Convert the audio
				SDL_ConvertAudio(&current_audio->convert);

				// Copy from SDL buffer to DMA buffer
				memcpy(dma_buffers[whichab], current_audio->convert.buf, current_audio->convert.len_cvt);
				dma_buffers_size[whichab] = current_audio->convert.len_cvt&~31;
			} else {
				current_audio->spec.callback(
					current_audio->spec.userdata,
					(Uint8 *)dma_buffers[whichab],
					DMA_BUFFER_SIZE);
			}

			SDL_UnlockMutex(current_audio->mixer_lock);
		}

		DCFlushRange(dma_buffers[whichab], dma_buffers_size[whichab]);

		AUDIO_InitDMA ((Uint32)dma_buffers[whichab], dma_buffers_size[whichab]);
		// Start the first chunk of audio playing
		if (!AUDIO_GetDMAEnableFlag())
			AUDIO_StartDMA();

		whichab ^= 1;
		LWP_SuspendThread(athread);
	}
	return NULL;
}

/****************************************************************************
 * DMACallback
 * signal audio thread that more samples are required
 ***************************************************************************/
static void
DMACallback()
{
	LWP_ResumeThread(athread);
}

static void WII_AudioStop()
{
	AUDIO_StopDMA();
	AUDIO_RegisterDMACallback(0);
	stopaudio = true;
	LWP_ResumeThread(athread);
	LWP_JoinThread(athread, NULL);
	athread = LWP_THREAD_NULL;
}

static void WII_AudioStart()
{
	AUDIO_SetDSPSampleRate(AI_SAMPLERATE_48KHZ);

	// startup conversion thread
	stopaudio = false;
	AUDIO_RegisterDMACallback(DMACallback);
	LWP_CreateThread(&athread, AudioThread, NULL, astack, AUDIOSTACK, 67);
}

static int WIIAUD_OpenAudio(_THIS, SDL_AudioSpec *spec)
{
	// Set up actual spec.
	spec->freq = 48000;
	spec->format	= AUDIO_S16MSB;
	spec->channels	= 2;
	spec->samples	= SAMPLES_PER_DMA_BUFFER;
	spec->padding	= 0;
	SDL_CalculateAudioSpec(spec);

	memset(dma_buffers[0], 0, sizeof(dma_buffers[0]));
	memset(dma_buffers[1], 0, sizeof(dma_buffers[0]));

	WII_AudioStart();

	return 1;
}

static void WIIAUD_WaitAudio(_THIS)
{

}

static void WIIAUD_PlayAudio(_THIS)
{

}

static Uint8 *WIIAUD_GetAudioBuf(_THIS)
{
	return NULL;
}

static void WIIAUD_CloseAudio(_THIS)
{
	WII_AudioStop();
}

static void WIIAUD_DeleteDevice(SDL_AudioDevice *device)
{
	WII_AudioStop();

	SDL_free(device->hidden);
	SDL_free(device);
}

static SDL_AudioDevice *WIIAUD_CreateDevice(int devindex)
{
	SDL_AudioDevice *this;

	/* Initialize all variables that we clean on shutdown */
	this = (SDL_AudioDevice *)SDL_malloc(sizeof(SDL_AudioDevice));
	if ( this ) {
		SDL_memset(this, 0, (sizeof *this));
		this->hidden = (struct SDL_PrivateAudioData *)
				SDL_malloc((sizeof *this->hidden));
	}
	if ( (this == NULL) || (this->hidden == NULL) ) {
		SDL_OutOfMemory();
		if ( this ) {
			SDL_free(this);
		}
		return(0);
	}
	SDL_memset(this->hidden, 0, (sizeof *this->hidden));

	// Initialise the Wii side of the audio system
	AUDIO_Init(NULL);

	/* Set the function pointers */
	this->OpenAudio = WIIAUD_OpenAudio;
	this->WaitAudio = WIIAUD_WaitAudio;
	this->PlayAudio = WIIAUD_PlayAudio;
	this->GetAudioBuf = WIIAUD_GetAudioBuf;
	this->CloseAudio = WIIAUD_CloseAudio;
	this->free = WIIAUD_DeleteDevice;

	return this;
}

static int WIIAUD_Available(void)
{
	return 1;
}

AudioBootStrap WIIAUD_bootstrap = {
	WIIAUD_DRIVER_NAME, "SDL Wii audio driver",
	WIIAUD_Available, WIIAUD_CreateDevice
};
