/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997-2006 Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


    Yohanes Nugroho (yohanes@gmail.com)
    Tantric, Extrems
*/
#include "SDL_config.h"

#include "../../events/SDL_sysevents.h"
#include "../../events/SDL_events_c.h"
#include <ogc/pad.h>
#include <ogc/usbmouse.h>
#include <wiikeyboard/keyboard.h>
#include <wiiuse/wpad.h>

#include "SDL_wiivideo.h"
#include "SDL_wiievents_c.h"

typedef struct {
	Uint32 mask;
	Uint16 sym;
	Uint16 mod;
} PadMap;

static const PadMap gamecube_keys[] = {
	{ PAD_BUTTON_UP,      SDLK_UP,       KMOD_NONE },
	{ PAD_BUTTON_DOWN,    SDLK_DOWN,     KMOD_NONE },
	{ PAD_BUTTON_LEFT,    SDLK_LEFT,     KMOD_NONE },
	{ PAD_BUTTON_RIGHT,   SDLK_RIGHT,    KMOD_NONE },
	{ PAD_BUTTON_START,   SDLK_q,        KMOD_CTRL },
	{ PAD_TRIGGER_L,      SDLK_KP_MINUS, KMOD_NONE },
	{ PAD_TRIGGER_R,      SDLK_KP_PLUS,  KMOD_NONE },
	{ PAD_BUTTON_X,       SDLK_DELETE,   KMOD_NONE },
	{ PAD_BUTTON_Y,       SDLK_x,        KMOD_NONE },
	{ PAD_TRIGGER_Z,      SDLK_z,        KMOD_NONE },
	{ PAD_SUBSTICK_LEFT,  SDLK_PAUSE,    KMOD_NONE },
	{ PAD_SUBSTICK_RIGHT, SDLK_TAB,      KMOD_NONE }
};

static const PadMap wiimote_keys[] = {
	{ WPAD_BUTTON_UP,    SDLK_UP,       KMOD_NONE },
	{ WPAD_BUTTON_DOWN,  SDLK_DOWN,     KMOD_NONE },
	{ WPAD_BUTTON_LEFT,  SDLK_LEFT,     KMOD_NONE },
	{ WPAD_BUTTON_RIGHT, SDLK_RIGHT,    KMOD_NONE },
	{ WPAD_BUTTON_HOME,  SDLK_q,        KMOD_CTRL },
	{ WPAD_BUTTON_MINUS, SDLK_KP_MINUS, KMOD_NONE },
	{ WPAD_BUTTON_PLUS,  SDLK_KP_PLUS,  KMOD_NONE },
	{ WPAD_BUTTON_1,     SDLK_DELETE,   KMOD_NONE },
	{ WPAD_BUTTON_2,     SDLK_x,        KMOD_NONE }
};

static const PadMap nunchuk_keys[] = {
	{ WPAD_NUNCHUK_BUTTON_C, SDLK_c, KMOD_NONE },
	{ WPAD_NUNCHUK_BUTTON_Z, SDLK_z, KMOD_NONE }
};

static const PadMap classic_keys[] = {
	{ WPAD_CLASSIC_BUTTON_UP,     SDLK_UP,       KMOD_NONE },
	{ WPAD_CLASSIC_BUTTON_DOWN,   SDLK_DOWN,     KMOD_NONE },
	{ WPAD_CLASSIC_BUTTON_LEFT,   SDLK_LEFT,     KMOD_NONE },
	{ WPAD_CLASSIC_BUTTON_RIGHT,  SDLK_RIGHT,    KMOD_NONE },
	{ WPAD_CLASSIC_BUTTON_HOME,   SDLK_q,        KMOD_CTRL },
	{ WPAD_CLASSIC_BUTTON_MINUS,  SDLK_KP_MINUS, KMOD_NONE },
	{ WPAD_CLASSIC_BUTTON_PLUS,   SDLK_KP_PLUS,  KMOD_NONE },
	{ WPAD_CLASSIC_BUTTON_X,      SDLK_DELETE,   KMOD_NONE },
	{ WPAD_CLASSIC_BUTTON_Y,      SDLK_x,        KMOD_NONE },
	{ WPAD_CLASSIC_BUTTON_FULL_L, SDLK_c,        KMOD_NONE },
	{ WPAD_CLASSIC_BUTTON_FULL_R, SDLK_z,        KMOD_NONE },
	{ WPAD_CLASSIC_BUTTON_ZL,     SDLK_PAUSE,    KMOD_NONE },
	{ WPAD_CLASSIC_BUTTON_ZR,     SDLK_TAB,      KMOD_NONE }
};

static SDLKey keymap[232];
static int posted;

extern bool TerminateRequested;
extern void Terminate();

static void handle_gamepad(_THIS)
{
	int rx, ry, i;
	
	if (PAD_ScanPads() & 1) {
		Uint32 pressed = PAD_ButtonsDown(PAD_CHAN0);
		Uint32 released = PAD_ButtonsUp(PAD_CHAN0);
		
		rx = PAD_StickX(PAD_CHAN0) / 8;
		ry = PAD_StickY(PAD_CHAN0) / 8;
		
		if (rx || ry)
			posted += SDL_PrivateMouseMotion(0, 1, rx, -ry);
		
		for (i = 0; i < SDL_arraysize(gamecube_keys); i++) {
			if (pressed & gamecube_keys[i].mask)
				posted += SDL_PrivateKeyboard2(SDL_PRESSED, gamecube_keys[i].sym, gamecube_keys[i].mod);
			else if (released & gamecube_keys[i].mask)
				posted += SDL_PrivateKeyboard2(SDL_RELEASED, gamecube_keys[i].sym, gamecube_keys[i].mod);
		}
		
		if (pressed & PAD_BUTTON_A)
			posted += SDL_PrivateMouseButton(SDL_PRESSED, SDL_BUTTON_LEFT, 0, 0);
		else if (released & PAD_BUTTON_A)
			posted += SDL_PrivateMouseButton(SDL_RELEASED, SDL_BUTTON_LEFT, 0, 0);
		
		if (pressed & PAD_BUTTON_B)
			posted += SDL_PrivateMouseButton(SDL_PRESSED, SDL_BUTTON_RIGHT, 0, 0);
		else if (released & PAD_BUTTON_B)
			posted += SDL_PrivateMouseButton(SDL_RELEASED, SDL_BUTTON_RIGHT, 0, 0);
	}
	
	if (WPAD_ReadPending(WPAD_CHAN_0, NULL) > 0) {
		WPADData *wd = WPAD_Data(WPAD_CHAN_0);
		
		if (wd->ir.valid) {
			posted += SDL_PrivateMouseMotion(0, 0, wd->ir.x - 48, wd->ir.y - 48);
		} else {
			if (wd->exp.type == WPAD_EXP_NUNCHUK) {
				nunchuk_t nunchuk = wd->exp.nunchuk;
				
				rx = (nunchuk.js.pos.x - nunchuk.js.center.x) * 32 / nunchuk.js.max.x;
				ry = (nunchuk.js.pos.y - nunchuk.js.center.y) * 32 / nunchuk.js.max.y;
			} else if (wd->exp.type == WPAD_EXP_CLASSIC) {
				classic_ctrl_t classic = wd->exp.classic;
				
				if ((classic.ljs.pos.x ^ classic.ljs.center.x) < 4)
					classic.ljs.pos.x = classic.ljs.center.x;
				if ((classic.rjs.pos.x ^ classic.rjs.center.x) < 4)
					classic.rjs.pos.x = classic.rjs.center.x;
				
				rx = ((classic.ljs.pos.x - classic.ljs.center.x) * 32 / classic.ljs.max.x) +
				     ((classic.rjs.pos.x - classic.rjs.center.x) * 32 / classic.rjs.max.x);
				
				if ((classic.ljs.pos.y ^ classic.ljs.center.y) < 4)
					classic.ljs.pos.y = classic.ljs.center.y;
				if ((classic.rjs.pos.y ^ classic.rjs.center.y) < 4)
					classic.rjs.pos.y = classic.rjs.center.y;
				
				ry = ((classic.ljs.pos.y - classic.ljs.center.y) * 32 / classic.ljs.max.y) +
				     ((classic.rjs.pos.y - classic.rjs.center.y) * 32 / classic.rjs.max.y);
			} else {
				rx = 0, ry = 0;
			}
			
			if (rx || ry)
				posted += SDL_PrivateMouseMotion(0, 1, rx, -ry);
		}
		
		for (i = 0; i < SDL_arraysize(wiimote_keys); i++) {
			if (wd->btns_d & wiimote_keys[i].mask)
				posted += SDL_PrivateKeyboard2(SDL_PRESSED, wiimote_keys[i].sym, wiimote_keys[i].mod);
			else if (wd->btns_u & wiimote_keys[i].mask)
				posted += SDL_PrivateKeyboard2(SDL_RELEASED, wiimote_keys[i].sym, wiimote_keys[i].mod);
		}
		
		if (wd->btns_d & WPAD_BUTTON_A)
			posted += SDL_PrivateMouseButton(SDL_PRESSED, SDL_BUTTON_LEFT, 0, 0);
		else if (wd->btns_u & WPAD_BUTTON_A)
			posted += SDL_PrivateMouseButton(SDL_RELEASED, SDL_BUTTON_LEFT, 0, 0);
		
		if (wd->btns_d & WPAD_BUTTON_B)
			posted += SDL_PrivateMouseButton(SDL_PRESSED, SDL_BUTTON_RIGHT, 0, 0);
		else if (wd->btns_u & WPAD_BUTTON_B)
			posted += SDL_PrivateMouseButton(SDL_RELEASED, SDL_BUTTON_RIGHT, 0, 0);
		
		if (wd->exp.type == WPAD_EXP_NUNCHUK) {
			for (i = 0; i < SDL_arraysize(nunchuk_keys); i++) {
				if (wd->btns_d & nunchuk_keys[i].mask)
					posted += SDL_PrivateKeyboard2(SDL_PRESSED, nunchuk_keys[i].sym, nunchuk_keys[i].mod);
				else if (wd->btns_u & nunchuk_keys[i].mask)
					posted += SDL_PrivateKeyboard2(SDL_RELEASED, nunchuk_keys[i].sym, nunchuk_keys[i].mod);
			}
		} else if (wd->exp.type == WPAD_EXP_CLASSIC) {
			for (i = 0; i < SDL_arraysize(classic_keys); i++) {
				if (wd->btns_d & classic_keys[i].mask)
					posted += SDL_PrivateKeyboard2(SDL_PRESSED, classic_keys[i].sym, classic_keys[i].mod);
				else if (wd->btns_u & classic_keys[i].mask)
					posted += SDL_PrivateKeyboard2(SDL_RELEASED, classic_keys[i].sym, classic_keys[i].mod);
			}
			
			if (wd->btns_d & WPAD_CLASSIC_BUTTON_A)
				posted += SDL_PrivateMouseButton(SDL_PRESSED, SDL_BUTTON_LEFT, 0, 0);
			else if (wd->btns_u & WPAD_CLASSIC_BUTTON_A)
				posted += SDL_PrivateMouseButton(SDL_RELEASED, SDL_BUTTON_LEFT, 0, 0);
			
			if (wd->btns_d & WPAD_CLASSIC_BUTTON_B)
				posted += SDL_PrivateMouseButton(SDL_PRESSED, SDL_BUTTON_RIGHT, 0, 0);
			else if (wd->btns_u & WPAD_CLASSIC_BUTTON_B)
				posted += SDL_PrivateMouseButton(SDL_RELEASED, SDL_BUTTON_RIGHT, 0, 0);
		}
	}
}

static void handle_keyboard(_THIS)
{
	keyboard_event ke;
	SDL_keysym keysym;
	
	if (KEYBOARD_GetEvent(&ke)) {
		if (ke.type == KEYBOARD_PRESSED || ke.type == KEYBOARD_RELEASED) {
			keysym.scancode = ke.keycode;
			keysym.sym = keymap[ke.keycode];
			keysym.mod = KMOD_NONE;
			keysym.unicode = ke.symbol;
			
			posted += SDL_PrivateKeyboard((ke.type == KEYBOARD_PRESSED) ? SDL_PRESSED : SDL_RELEASED, &keysym);
		}
	}
}

static void handle_mouse(_THIS)
{
	mouse_event me;
	Uint8 state;
	
	if (MOUSE_GetEvent(&me)) {
		state = SDL_GetMouseState(NULL, NULL);
		
		if (me.rx || me.ry)
			posted += SDL_PrivateMouseMotion(0, 1, me.rx, me.ry);
		
		if (me.rz > 0) {
			posted += SDL_PrivateMouseButton(SDL_PRESSED, SDL_BUTTON_WHEELUP, 0, 0);
		} else {
			if (state & SDL_BUTTON(SDL_BUTTON_WHEELUP))
				posted += SDL_PrivateMouseButton(SDL_RELEASED, SDL_BUTTON_WHEELUP, 0, 0);
		}
		
		if (me.rz < 0) {
			posted += SDL_PrivateMouseButton(SDL_PRESSED, SDL_BUTTON_WHEELDOWN, 0, 0);
		} else {
			if (state & SDL_BUTTON(SDL_BUTTON_WHEELDOWN))
				posted += SDL_PrivateMouseButton(SDL_RELEASED, SDL_BUTTON_WHEELDOWN, 0, 0);
		}
		
		if (me.button & 0x1) {
			if (!(state & SDL_BUTTON_LMASK))
				posted += SDL_PrivateMouseButton(SDL_PRESSED, SDL_BUTTON_LEFT, 0, 0);
		} else {
			if (state & SDL_BUTTON_LMASK)
				posted += SDL_PrivateMouseButton(SDL_RELEASED, SDL_BUTTON_LEFT, 0, 0);
		}
		
		if (me.button & 0x2) {
			if (!(state & SDL_BUTTON_RMASK))
				posted += SDL_PrivateMouseButton(SDL_PRESSED, SDL_BUTTON_RIGHT, 0, 0);
		} else {
			if (state & SDL_BUTTON_RMASK)
				posted += SDL_PrivateMouseButton(SDL_RELEASED, SDL_BUTTON_RIGHT, 0, 0);
		}
		
		if (me.button & 0x4) {
			if (!(state & SDL_BUTTON_MMASK))
				posted += SDL_PrivateMouseButton(SDL_PRESSED, SDL_BUTTON_MIDDLE, 0, 0);
		} else {
			if (state & SDL_BUTTON_MMASK)
				posted += SDL_PrivateMouseButton(SDL_RELEASED, SDL_BUTTON_MIDDLE, 0, 0);
		}
	}
}

void WII_PumpEvents(_THIS)
{
	do {
		if (TerminateRequested)
			Terminate();
		posted = 0;
		handle_gamepad(this);
		handle_keyboard(this);
		handle_mouse(this);
	} while (posted);
}

void WII_InitOSKeymap(_THIS)
{
	int i;
	
	for (i = 0; i < SDL_arraysize(keymap); i++)
		keymap[i] = SDLK_UNKNOWN;
	
	// a-z
	for (i = 0; i < 27; i++)
		keymap[4 + i] = SDLK_a + i;
	
	// 1-9
	for (i = 0; i < 9; i++)
		keymap[30 + i] = SDLK_1 + i;
	
	keymap[39] = SDLK_0;
	keymap[40] = SDLK_RETURN;
	keymap[41] = SDLK_ESCAPE;
	keymap[42] = SDLK_BACKSPACE;
	keymap[43] = SDLK_TAB;
	keymap[44] = SDLK_SPACE;
	keymap[45] = SDLK_MINUS;
	keymap[46] = SDLK_EQUALS;
	keymap[47] = SDLK_LEFTBRACKET;
	keymap[48] = SDLK_RIGHTBRACKET;
	keymap[49] = SDLK_BACKSLASH;
	keymap[50] = SDLK_UNKNOWN; // unused
	keymap[51] = SDLK_SEMICOLON;
	keymap[52] = SDLK_QUOTE;
	keymap[53] = SDLK_BACKQUOTE;
	keymap[54] = SDLK_COMMA;
	keymap[55] = SDLK_PERIOD;
	keymap[56] = SDLK_SLASH;
	keymap[57] = SDLK_CAPSLOCK;
	
	// F1 to F12
	for (i = 0; i < 12; i++)
		keymap[58 + i] = SDLK_F1 + i;
	
	keymap[70] = SDLK_PRINT;
	keymap[71] = SDLK_SCROLLOCK;
	keymap[72] = SDLK_PAUSE;
	keymap[73] = SDLK_INSERT;
	keymap[74] = SDLK_HOME;
	keymap[75] = SDLK_PAGEUP;
	keymap[76] = SDLK_DELETE;
	keymap[77] = SDLK_END;
	keymap[78] = SDLK_PAGEDOWN;
	
	keymap[79] = SDLK_RIGHT;
	keymap[80] = SDLK_LEFT;
	keymap[81] = SDLK_DOWN;
	keymap[82] = SDLK_UP;
	keymap[83] = SDLK_NUMLOCK;
	
	keymap[84] = SDLK_KP_DIVIDE;
	keymap[85] = SDLK_KP_MULTIPLY;
	keymap[86] = SDLK_KP_MINUS;
	keymap[87] = SDLK_KP_PLUS;
	
	keymap[88] = SDLK_KP_ENTER;
	
	// keypad 1-9
	for (i = 0; i < 9; i++)
		keymap[89 + i] = SDLK_KP1 + i;
	
	keymap[98] = SDLK_KP0;
	keymap[99] = SDLK_KP_PERIOD;
	keymap[100] = SDLK_UNKNOWN; // unused
	keymap[101] = SDLK_UNKNOWN; // Applic
	keymap[102] = SDLK_POWER;
	keymap[103] = SDLK_KP_EQUALS;
	keymap[104] = SDLK_F13;
	keymap[105] = SDLK_F14;
	keymap[106] = SDLK_F15;
	keymap[107] = SDLK_UNKNOWN; // F16
	keymap[108] = SDLK_UNKNOWN; // F17
	keymap[109] = SDLK_UNKNOWN; // F18
	keymap[110] = SDLK_UNKNOWN; // F19
	keymap[111] = SDLK_UNKNOWN; // F20
	keymap[112] = SDLK_UNKNOWN; // F21
	keymap[113] = SDLK_UNKNOWN; // F22
	keymap[114] = SDLK_UNKNOWN; // F23
	keymap[115] = SDLK_UNKNOWN; // F24
	keymap[116] = SDLK_UNKNOWN; // Execute
	keymap[117] = SDLK_HELP;
	keymap[118] = SDLK_MENU;
	keymap[119] = SDLK_UNKNOWN; // Select
	keymap[120] = SDLK_UNKNOWN; // Stop
	keymap[121] = SDLK_UNKNOWN; // Again
	keymap[122] = SDLK_UNDO;
	
	keymap[134] = SDLK_KP_EQUALS;
	// 135-143 - International
	// 144-152 - Language
	keymap[154] = SDLK_SYSREQ;
	
	keymap[224] = SDLK_LCTRL;
	keymap[225] = SDLK_LSHIFT;
	keymap[226] = SDLK_LALT;
	keymap[227] = SDLK_LMETA;
	keymap[228] = SDLK_RCTRL;
	keymap[229] = SDLK_RSHIFT;
	keymap[230] = SDLK_RALT;
	keymap[231] = SDLK_RMETA;
}
