/* $Id$ */

/*
 * This file is part of OpenTTD.
 * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
 * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
 */
 
#include "../stdafx.h"
#include "thread.h"
#include <ogc/lwp.h>
#include <ogc/mutex.h>
#include <ogc/cond.h>

/**
 * libOGC version for ThreadObject.
 */
class ThreadObject_LWP : public ThreadObject {
private:
	lwp_t thread;        ///< System thread identifier.
	OTTDThreadFunc proc; ///< External thread procedure.
	void *param;         ///< Parameter for the external thread procedure.
	bool self_destruct;  ///< Free ourselves when done?

public:
	/**
	 * Create a thread and start it, calling proc(param).
	 */
	ThreadObject_LWP(OTTDThreadFunc proc, void *param, bool self_destruct) :
		thread(LWP_THREAD_NULL),
		proc(proc),
		param(param),
		self_destruct(self_destruct)
	{
		LWP_CreateThread(&this->thread, &stThreadProc, this, NULL, 32768, LWP_PRIO_NORMAL);
	}

	/* virtual*/  bool Exit()
	{
		assert(LWP_GetSelf() == this->thread);
		/* For now we terminate by throwing an error, gives much cleaner cleanup */
		throw OTTDThreadExitSignal();
	}

	/* virtual */ void Join()
	{
		/* You cannot join yourself */
		assert(LWP_GetSelf() != this->thread);
		LWP_JoinThread(this->thread, NULL);
		this->thread = LWP_THREAD_NULL;
	}

private:
	/**
	 * On thread creation, this function is called, which calls the real startup
	 *  function. This to get back into the correct instance again.
	 */
	static void *stThreadProc(void *thr)
	{
		((ThreadObject_LWP *)thr)->ThreadProc();
		return NULL;
	}

	/**
	 * A new thread is created, and this function is called. Call the custom
	 *  function of the creator of the thread.
	 */
	void ThreadProc()
	{
		/* Call the proc of the creator to continue this thread */
		try {
			this->proc(this->param);
		} catch (OTTDThreadExitSignal e) {
		} catch (...) {
			NOT_REACHED();
		}

		if (self_destruct) delete this;
	}
};

/* static */ bool ThreadObject::New(OTTDThreadFunc proc, void *param, ThreadObject **thread)
{
	ThreadObject *to = new ThreadObject_LWP(proc, param, thread == NULL);
	if (thread != NULL) *thread = to;
	return true;
}

/**
 * libOGC version of ThreadMutex.
 */
class ThreadMutex_LWP : public ThreadMutex {
private:
	mutex_t mutex;
	cond_t condition;

public:
	ThreadMutex_LWP()
	{
		LWP_MutexInit(&this->mutex, false);
		LWP_CondInit(&this->condition);
	}

	/* virtual */ ~ThreadMutex_LWP()
	{
		LWP_MutexDestroy(this->mutex);
		LWP_CondDestroy(this->condition);
	}

	/* virtual */ void BeginCritical()
	{
		LWP_MutexLock(this->mutex);
	}

	/* virtual */ void EndCritical()
	{
		LWP_MutexUnlock(this->mutex);
	}
	
	/* virtual */ void WaitForSignal()
	{
		LWP_CondWait(this->condition, this->mutex);
	}

	/* virtual */ void SendSignal()
	{
		LWP_CondSignal(this->condition);
	}
};

/* static */ ThreadMutex *ThreadMutex::New()
{
	return new ThreadMutex_LWP();
}
